//
// Gaggle models
// by Jarno van der Linden  jvan006@cs.auckland.ac.nz
//

#include "gaggle.h"

#include <fstream>

using namespace Gaggle;
using namespace std;


//
// The ModelData superclass is used for models and textures
// It keeps the raw (compressed) data in memory to send to
// Worlds or Players when requested, without having to load
// it from disk every time
//

ModelData::ModelData()
	: Identified()
{
}

ModelData::ModelData(const unsigned char *data, size_t datasize)
	: Identified()
{
	AddModelData(data, datasize);
}

ModelData::ModelData(const string& filename)
	: Identified()
{
	LoadData(filename);
}

ModelData::ModelData(const ModelDataMessage& msg, size_t offset)
	: Identified(msg)
{
	AddModelData(((unsigned char *)&msg)+offset, msg.compressedsize);
}

ModelData::~ModelData()
{
}


void ModelData::AddModelData(const unsigned char* data, size_t datasize)
{
	size_t cursize = compresseddata.size();
	compresseddata.resize(cursize + datasize);
	unsigned char *ptr = &(compresseddata[cursize]);
	memcpy(ptr, data, datasize);
}


bool ModelData::LoadData(const string& filename)
{
	compresseddata.clear();

	ifstream is;
	is.open(filename.c_str(), ios_base::binary);
	if(is.fail())
		return false;

	is.seekg(0, ios_base::end);
	streampos end = is.tellg();
	is.seekg(0, ios_base::beg);

	compresseddata.resize(end);

	is.read((char *)&compresseddata[0], end);

	return true;
}


size_t ModelData::GetCompressedSize(void) const
{
	return compresseddata.size();
}


const unsigned char *ModelData::GetCompressedData(void) const
{
	if(compresseddata.empty())
		return NULL;

	return &(compresseddata[0]);
}


// A message containing some model data
ModelDataMessage* ModelDataMessage::Create(const ModelData& modeldata, unsigned char typeID, unsigned int size)
{
	unsigned int modeldatamsgsize = size + modeldata.GetCompressedSize();
	ModelDataMessage *modeldatamsg = (ModelDataMessage *)new unsigned char[modeldatamsgsize];

	IdentifiedMessage::Fill(*modeldatamsg, modeldata, typeID, modeldatamsgsize);
	modeldatamsg->compressedsize = modeldata.GetCompressedSize();
	memcpy(((unsigned char *)modeldatamsg)+size, modeldata.GetCompressedData(), modeldatamsg->compressedsize);

	return modeldatamsg;
}

void ModelDataMessage::Destroy(ModelDataMessage* msg)
{
	delete[] msg;
}


Model::Model(int modeltype_in)
	: ModelData(), modeltype(modeltype_in)
{
}

Model::Model(unsigned char* data, size_t datasize, int modeltype_in)
	: ModelData(data, datasize), modeltype(modeltype_in)
{
}

Model::Model(const string& filename, int modeltype_in)
	: ModelData(filename), modeltype(modeltype_in)
{
}

Model::Model(const ModelMessage& msg, size_t offset)
	: ModelData(msg, offset), modeltype(msg.modeltype)
{
}

Model::~Model()
{
}


int Model::GetModelType(void) const
{
	return modeltype;
}


void Model::Render(void)
{
}


// A message containing a model
ModelMessage* ModelMessage::Create(const Model& model)
{
	ModelMessage *modelmsg = (ModelMessage *)ModelDataMessage::Create(model, ID_MODEL_MESSAGE, sizeof(ModelMessage));
	modelmsg->modeltype = model.GetModelType();
	return modelmsg;
}

// A message requesting a model
void AskModelMessage::Fill(AskModelMessage& msg, UID uid)
{
	GaggleMessage::Fill(msg, ID_ASKMODEL_MESSAGE, sizeof(AskModelMessage));
	msg.uid = uid;
}


//
// A Static Model is composed of arrays of vertices and texture
// coordinates. Geometry is created using the given OpenGL primitive
// type (GL_TRIANGLES, GL_QUADSTRIP, etc.) and vertex index array.
//
unsigned int StaticModel::magic = 0x4747534d;

StaticModel::StaticModel(const Vector* verts_in, const UV* texuv_in, unsigned int numverts_in, const int* idx_in, unsigned int numidx_in, unsigned int primtype_in)
	: Model(MODELTYPE_STATIC), numverts(numverts_in), numidx(numidx_in), primtype(primtype_in)
{
	AddModelData((unsigned char *)&magic, sizeof(magic));
	AddModelData((unsigned char *)&primtype, sizeof(primtype));
	AddModelData((unsigned char *)&numverts, sizeof(numverts));
	AddModelData((unsigned char *)&numidx, sizeof(numidx));
	AddModelData((unsigned char *)verts_in, numverts * sizeof(*verts_in));
	AddModelData((unsigned char *)texuv_in, numverts * sizeof(*texuv_in));
	AddModelData((unsigned char *)idx_in, numidx * sizeof(*idx_in));

	verts = (const Vector *)&(compresseddata[0]);
	texuv = (const UV *)(verts + numverts);
	idx = (const int *)(texuv + numverts);
}

StaticModel::StaticModel(const ModelMessage& msg)
	: Model(msg, sizeof(msg)), numverts(0), numidx(0), verts(0), texuv(0), idx(0), primtype(0)
{
	DecompressFromMem(GetCompressedData(), GetCompressedSize());
}

StaticModel::~StaticModel()
{
}

bool StaticModel::DecompressFromMem(const unsigned char* data, const size_t datasize)
{
	const unsigned char *ptr = data;
	size_t dataleft = datasize;

	primtype = 0;
	numverts = numidx = 0;
	verts = NULL;
	texuv = NULL;
	idx = NULL;

	if(dataleft < (sizeof(magic) + sizeof(primtype) + sizeof(numverts) + sizeof(numidx)))
		return false;

	if((*(unsigned int *)ptr) != magic)
		return false;

	ptr += sizeof(magic); dataleft -= sizeof(magic);
	primtype = *(unsigned int *)ptr;
	ptr += sizeof(primtype); dataleft -= sizeof(primtype);
	numverts = *(unsigned int *)ptr;
	ptr += sizeof(numverts); dataleft -= sizeof(numverts);
	numidx = *(unsigned int *)ptr;
	ptr += sizeof(numidx); dataleft -= sizeof(numidx);
	if(dataleft < (sizeof(*verts) * numverts + sizeof(*texuv) * numverts + sizeof(*idx) * numidx))
	{
		primtype = 0;
		numverts = numidx = 0;
		return false;
	}

	verts = (const Vector *)ptr;
	ptr += sizeof(*verts) * numverts; dataleft -= sizeof(*verts) * numverts;
	texuv = (const UV *)ptr;
	ptr += sizeof(*texuv) * numverts; dataleft -= sizeof(*texuv) * numverts;
	idx = (const int *)ptr;

	return true;
}


//
// An MD2 model, which may be animated
// The MD2 model is composed of a number of frames.
// For each frame, an array of vertex positions is stored.
// Each frame has the same number of vertices.
// Vertices have UV coordinates which are the same for all
// frames, and are stored in an array.
// There are two ways of rendering an MD2 model.
// The first is to render each triangle using GL_TRIANGLES.
// Indices into the vertex and UV arrays for each triangle
// are stored in an array.
// The second way to render an MD2 model is to use the
// glcommands array. The glcommands array is a sequence of
// triangle strips or triangle fans. A single command consists
// of a number of vertices with UV coordinates and indices into
// the verts array. In pseudo-C++ a single command looks like:
//   struct GLCommand
//   {
//     int numverts;
//     VertexData vertdata[abs(numverts)];
//   };
//   struct VertexData
//   {
//     float u, v;
//     int vertidx;
//   };
// If numverts is larger than 0, the vertices form a triangle
// strip. If numverts is less than 0, the vertices form a
// triangle fan. The end of the glcommands array is indicated
// by numverts being equal to 0.
// For an example of this, look at rendermd2.cpp of the Cube
// game source code available through the 777 lectures webpage.
//

MD2Model::MD2Model(const string& filename)
	: Model(filename, MODELTYPE_MD2)
{
	DecompressFromMem(GetCompressedData(), GetCompressedSize());
}

MD2Model::MD2Model(const ModelMessage& msg)
	: Model(msg, sizeof(msg))
{
	DecompressFromMem(GetCompressedData(), GetCompressedSize());
}

MD2Model::~MD2Model()
{
}


//
// MD2 parsing based on the MD2 animation tutorial by Digiben
// digiben@gametutorials.com   http://www.gametutorials.com/
//

// These are the needed defines for the max values when loading .MD2 files
#define MD2_MAX_TRIANGLES		4096
#define MD2_MAX_VERTICES		2048
#define MD2_MAX_TEXCOORDS		2048
#define MD2_MAX_FRAMES			512
#define MD2_MAX_SKINS			32
#define MD2_MAX_FRAMESIZE		(MD2_MAX_VERTICES * 4 + 128)

// This stores the speed of the animation between each key frame
#define kAnimationSpeed			5.0f

// This holds the header information that is read in at the beginning of the file
struct tMd2Header
{ 
   int magic;					// This is used to identify the file
   int version;					// The version number of the file (Must be 8)
   int skinWidth;				// The skin width in pixels
   int skinHeight;				// The skin height in pixels
   int frameSize;				// The size in bytes the frames are
   int numSkins;				// The number of skins associated with the model
   int numVertices;				// The number of vertices (constant for each frame)
   int numTexCoords;			// The number of texture coordinates
   int numTriangles;			// The number of faces (polygons)
   int numGlCommands;			// The number of gl commands
   int numFrames;				// The number of animation frames
   int offsetSkins;				// The offset in the file for the skin data
   int offsetTexCoords;			// The offset in the file for the texture data
   int offsetTriangles;			// The offset in the file for the face data
   int offsetFrames;			// The offset in the file for the frames data
   int offsetGlCommands;		// The offset in the file for the gl commands data
   int offsetEnd;				// The end of the file offset
};

// This stores UV coordinates
struct tMd2TexCoord
{
   short u, v;
};

// This stores the indices into the vertex and texture coordinate arrays
struct tMd2Face
{
   short vertexIndices[3];
   short textureIndices[3];
};

// This is used to store the vertices that are read in for the current frame
struct tMd2AliasTriangle
{
   unsigned char vertex[3];
   unsigned char lightNormalIndex;
};

// This stores the animation scale, translation and name information for a frame, plus verts
struct tMd2AliasFrame
{
   float scale[3];
   float translate[3];
   char name[16];
   tMd2AliasTriangle aliasVertices[1];
};

// This stores the normals and vertices for the frames
struct tMd2Triangle
{
   float vertex[3];
   float normal[3];
};

// This stores the frames vertices after they have been transformed
struct tMd2Frame
{
	std::string strName;
   tMd2Triangle *pVertices;
};

bool MD2Model::DecompressFromMem(const unsigned char* data, const size_t datasize)
{
	tMd2Header* md2header;
	tMd2TexCoord* md2texcoords;
	tMd2Face* md2tris;
	int* md2glcommands;

	if(datasize < sizeof(*md2header))
		return false;

	md2header = (tMd2Header *)data;
	if(md2header->version != 8)
		return false;

	if((size_t)(md2header->offsetEnd) > datasize)
		return false;

	md2texcoords = (tMd2TexCoord *)(data + md2header->offsetTexCoords);
	md2tris = (tMd2Face *)(data + md2header->offsetTriangles);
	md2glcommands = (int *)(data + md2header->offsetGlCommands);

	texuv.reserve(md2header->numTexCoords);
	// Go through all of the uv coordinates and assign them over to our structure.
	// The UV coordinates are not normal uv coordinates, they have a pixel ratio of
	// 0 to 256.  We want it to be a 0 to 1 ratio, so we divide the u value by the
	// skin width and the v value by the skin height.  This gives us our 0 to 1 ratio.
	// For some reason also, the v coodinate is flipped upside down.  We just subtract
	// the v coordinate from 1 to remedy this problem.
	for(int i = 0; i < md2header->numTexCoords; i++)
	{
		UV uv = {
			md2texcoords[i].u / float(md2header->skinWidth),
			1 - md2texcoords[i].v / float(md2header->skinHeight)
		};
		texuv.push_back(uv);
	}

	triangles.reserve(md2header->numTriangles);
	for(int i = 0; i < md2header->numTriangles; i++)
	{
		const Triangle triangle = {
			md2tris[i].vertexIndices[0], md2tris[i].vertexIndices[1], md2tris[i].vertexIndices[2],
			md2tris[i].textureIndices[0], md2tris[i].textureIndices[1], md2tris[i].textureIndices[2]
		};
		triangles.push_back(triangle);
	}

	verts.resize(md2header->numFrames);

	for(int i = 0; i < md2header->numFrames; i++)
	{
		tMd2AliasFrame* frame = (tMd2AliasFrame *)(data + md2header->offsetFrames + i * md2header->frameSize);

		// Go through all of the number of vertices and assign the scale and translations.
		// Store the vertices in our current frame's vertex list array, while swapping Y and Z.
		// Notice we also negate the Z axis as well to make the swap correctly.
		verts[i].reserve(md2header->numVertices);
		for (int j = 0; j < md2header->numVertices; j++)
		{
			const Vector v = {
				frame->aliasVertices[j].vertex[0] * frame->scale[0] + frame->translate[0],
				frame->aliasVertices[j].vertex[2] * frame->scale[2] + frame->translate[2],
				-1 * (frame->aliasVertices[j].vertex[1] * frame->scale[1] + frame->translate[1])
			};
			verts[i].push_back(v);
		}
	}

	glcommands.reserve(md2header->numGlCommands);
	for(int i = 0; i < md2header->numGlCommands; i++)
	{
		glcommands.push_back(md2glcommands[i]);
	}

	return true;
}
